import numpy as np
import pandas as pd
from typing import Dict, Tuple

events_T = Dict[str, pd.DatetimeIndex]
mask_T = Dict[str, np.ndarray]
nested_T = Dict[str, pd.DataFrame]
outcomes_T = Dict[str, pd.Series]
pred_lcl_T = Dict[str, pd.Series]
pred_gbl_T = Dict[str, pd.DatetimeIndex]
prediction_T = Dict[str, Tuple[pd.Series, pd.DatetimeIndex]]
sensors_T = Dict[str, pd.DataFrame]



import pandas as pd
import pkg_resources
from typing import Dict, Tuple

def load_lunges() -> Tuple[Dict[str, pd.DataFrame], 
                           Dict[str, pd.DatetimeIndex]]:
    """Load sample data.

    Loads a small dataset containing six blue whale deployments and labeled 
    lunge-feeding behaviors.

    Returns:
        Tuple[Dict[str, pd.DataFrame], Dict[str, pd.DatetimeIndex]]: Sensors 
            and events, in Stickleback format.
    """
    sensors_file = (pkg_resources
                    .resource_filename('stickleback', 'data/sensors.csv'))
    sensors_csv = pd.read_csv(sensors_file)
    
    events_file = (pkg_resources
                   .resource_filename('stickleback', 'data/events.csv'))
    events_csv = pd.read_csv(events_file)
    
    sensors, events = dict(), dict()
    
    for k in sensors_csv["deployid"].unique():
        sensors[k] = (sensors_csv[sensors_csv["deployid"] == k]
                      .drop("deployid", axis=1)
                      .set_index("datetime"))
        sensors[k].index = pd.DatetimeIndex(sensors[k].index)
        
    for k in events_csv["deployid"].unique():
        events[k] = pd.DatetimeIndex(
            events_csv[events_csv["deployid"] == k]["datetime"]
        )
    
    return sensors, events



"""Automated behavioral event detection in bio-logging data with Stickleback.

Stickleback is a machine-learning pipeline for using time series classification
algorithms to detect fine-scale behavioral events in longitudinal bio-logging
data.

Exports class Stickleback, which has methods for model initializion, training,
prediction, and assessment.
"""

import scipy.signal as signal 
import sklearn
import sklearn.model_selection as selection 
import sktime.datatypes._panel._convert as convert 
import sktime.base
from stickleback.types import *
import stickleback.util as sb_util
import numpy as np
import pandas as pd

class Stickleback:
    """Automated behavioral event detection in bio-logging data.

    Stickleback applies arbitrary time series classification algorithms to
    longitudinal bio-logging data for detecting fine-scale behavioral events.
    It's compatible with the sktime package, which provides the time series
    classification algorithm implementations. The pipeline works by 1)
    extracting training data from labeled bio-logging data, 2) fitting the time
    series classifier, 3) choosing a probability threshold for event detection,
    and 4) boosting prediction accuracy. 
    For more information, see manuscript (in prep).
    """

    def __init__(self, 
                 local_clf: sktime.base.BaseEstimator, 
                 win_size: int, 
                 tol: pd.Timedelta, 
                 nth: int = 1, 
                 n_folds: int = 5, 
                 max_events: int = None,
                 seed: int = None) -> None:
        """Initialize a Stickleback object

        Args:
            local_clf (sktime.base.BaseEstimator): A time series classifier.
            win_size (int): Window size for sliding prediction window.
            tol (pd.Timedelta): Time tolerance for assessing prediction 
                accuracy.
            nth (int, optional): Sliding window step size. Defaults to 1.
            n_folds (int, optional): Number of folds for cross validation step.
                Defaults to 5.
            max_events (int, optional): Maximum number of labeled events to 
                include in model training. If None, uses all events. Defaults 
                to None.
            seed (int, optional): Seed for random number generation. Note: this 
                only applies to Stickleback methods; if local_clf uses random
                numbers then its seed needs to be set separately.
        """
        self.local_clf = local_clf
        self.__local_clf2 = sklearn.clone(local_clf)
        self.prominence = None
        self.win_size = win_size
        self.tol = tol
        self.nth = nth
        self.n_folds = n_folds
        self.max_events = max_events
        self.seed = seed

    def fit(self, 
            sensors: sensors_T, 
            events: events_T, 
            mask: mask_T = None) -> None:
        """Fits a Stickleback model.

        Args:
            sensors (Dict[str, pd.DataFrame]): Sensor data.
            events (Dict[str, pd.DatetimeIndex]): Labeled events.
            mask (Dict[str, np.ndarray], optional): Training mask. Only trains 
                model on windows in sensors[deployid] where mask[deployid] is 
                True. If None, no mask applied. Defaults to None.
        """
        # Local step
        if self.max_events is None:
            training_events = events
        else:
            n_events = np.array([len(v) for v in events.values()])
            if n_events.sum() <= self.max_events:
                training_events = events
            else:
                rg = np.random.Generator(np.random.PCG64(self.seed))
                keep = self.max_events / n_events.sum()
                training_events = {k: rg.choice(v, 
                                                size=int(len(v)*keep), 
                                                replace=False) 
                                   for k, v in events.items()}
                
        events_nested = sb_util.extract_nested(sensors, 
                                               training_events, 
                                               self.win_size)
        event_X = pd.concat(events_nested.values())
        nonevents_nested = sb_util.sample_nonevents(sensors, 
                                                    training_events, 
                                                    self.win_size,
                                                    mask,
                                                    self.seed)
        nonevent_X = pd.concat(nonevents_nested.values())
        local_X = event_X.append(nonevent_X)
        event_y = np.full(len(event_X), 1.0)
        nonevent_y = np.full(len(nonevent_X), 0.0)
        local_y = np.concatenate([event_y, nonevent_y])
        self._fit_local(local_X, local_y)

        # Global step (using internal cross validation)
        local_proba_cv = self._fit_global(events_nested, 
                                          nonevents_nested, 
                                          sensors, 
                                          events, 
                                          mask)

        # Boost
        predictions_gbl = self._predict_global(local_proba_cv)
        predictions = {k: (None, v) for k, v in predictions_gbl.items()}
        outcomes = self.assess(predictions, events)
        boosted_nonevents = self._boost(nonevents_nested, sensors, outcomes)
        n_nonevents = np.sum([len(v) for v in boosted_nonevents.values()])
        boosted_X = event_X.append(pd.concat(boosted_nonevents.values()))
        boosted_y = np.concatenate([event_y, np.full(n_nonevents, 0.0)])
        self._fit_local(boosted_X, boosted_y)
        self._fit_global(events_nested, 
                         boosted_nonevents, 
                         sensors, 
                         events, 
                         mask)

    def predict(self, 
                new_sensors: sensors_T, 
                new_mask: mask_T = None) -> prediction_T:
        """Make predictions with fitted Stickleback model.

        Args:
            new_sensors (Dict[str, pd.DataFrame]): New sensor data.
            new_mask (Dict[str, np.ndarray], optional): Prediction mask. Only 
                makes predictions for windows in new_sensors[deployid] where 
                mask[deployid] is True. If None, no mask applied. Defaults to 
                None.

        Returns:
            Dict[str, Tuple[pd.Series, pd.DatetimeIndex]]: Event predictions. 
                A dict keyed by deployment ID. Values are tuples: (local 
                probability, predicted event times).
        """
        local_proba = self._predict_local(new_sensors, new_mask)
        global_pred = self._predict_global(local_proba)
        return {d: (local_proba[d], global_pred[d]) for d in global_pred}

    def assess(self, predicted: prediction_T, events: events_T) -> outcomes_T:
        """Assess prediction accuracy

        Args:
            predicted (Dict[str, Tuple[pd.Series, pd.DatetimeIndex]]): Model 
                predictions, as from Stickleback.predict().
            events (Dict[str, pd.DatetimeIndex]): Labeled events.

        Returns:
            Dict[str, pd.Series]: A dict of prediction outcomes, keyed by 
                deployment ID. Values are Series made of "TP", "FP", and "FN" 
                (true positive, false positive, and false negative).
        """
        result = dict()
        pred_gbl = {k: v[1] for k, v in predicted.items()}
        for deployid in pred_gbl:
            _predicted, _actual = pred_gbl[deployid], events[deployid]
            # Handle edge cases for no predicted/actual events
            if len(_predicted) == 0 and len(_actual) > 0:
                return pd.Series("FN",
                                 index=_actual,
                                 dtype="string",
                                 name="outcome")
           
            if len(_predicted) > 0 and len(_actual) == 0:
                return pd.Series("FP",
                                 index=_predicted,
                                 dtype="string",
                                 name="outcome")
           
            if len(_predicted) == 0 and len(_actual) == 0:
                return pd.Series([],
                                 dtype="string",
                                 name="outcome")
            
            # Find closest predicted to each actual and their distance
            closest = _predicted[[np.argmin(np.abs(_predicted - e)) 
                                  for e in _actual]]
            distance = np.abs(_actual - closest)
            
            # Initialize outcomes
            outcomes = pd.Series(index=_predicted, 
                                 dtype="string", 
                                 name="outcome")

            # Iterate through actual events. The closest predicted event within
            # the tolerance is a true positive. If no predicted events are
            # within the tolerance, the actual event is a false negative.
            for i, (c, d) in enumerate(zip(closest, distance)):
                if d <= self.tol:
                    outcomes[c] = "TP" 
                else:
                    outcomes[_actual[i]] = "FN"

            # Remaining predictions are false positives
            outcomes[outcomes.isna()] = "FP"

            result[deployid] = outcomes
            
        return result

    def _fit_local(self, 
                   local_X: pd.DataFrame, 
                   local_y: np.ndarray, 
                   clone: bool = False) -> None:
        """Fit the local classifier.

        Args:
            local_X (pd.DataFrame): Training data (sensor windows, in sktime 
                nested DataFrame format).
            local_y (np.ndarray): Training labels (0's and 1's).
            clone (bool, optional): Clone flag. True for internal cross 
                validation. Defaults to False.
        """
        if len(local_X.columns) > 1:
            local_X = convert.from_nested_to_3d_numpy(local_X)

        clf = self.__local_clf2 if clone else self.local_clf
        clf.fit(local_X, local_y)

    def _predict_local(self, 
                       sensors: sensors_T, 
                       mask: mask_T = None,
                       clone: bool = False) -> pred_lcl_T:
        """Local step predictions.

        Args:
            sensors (Dict[str, pd.DataFrame]): Sensor data.
            mask (Dict[str, np.ndarray], optional): Prediction mask. Defaults 
                to None.
            clone (bool, optional): Clone flag. True for internal cross 
                validation. Defaults to False.

        Returns:
            Dict[str, pd.Series]: A dictionary of local predictions (i.e. 
                Series with local probability), keyed by deployment ID.
        """
        clf = self.local_clf if not clone else self.__local_clf2
        X = sb_util.extract_all(sensors, self.nth, self.win_size, mask)

        def __predict_local(_X: pd.DataFrame, 
                           i: pd.DatetimeIndex, 
                           _mask: np.ndarray):
            local = pd.Series(clf.predict_proba(_X)[:, 1], 
                              name="local_proba", 
                              index=_X.index). \
                    reindex(i)
            local[np.logical_not(_mask)] = 0.
            return local.interpolate(method="cubic").fillna(0)
        
        if mask is None:
            mask = {k: np.full(len(v), True) for k, v in sensors.items()}
        
        return {deployid: __predict_local(X[deployid],
                                         sensors[deployid].index, 
                                         mask[deployid]) 
                for deployid in X}

    def _fit_global(self, 
                    events_nested: nested_T, 
                    nonevents_nested: nested_T, 
                    sensors: sensors_T, 
                    events: events_T, 
                    mask: mask_T) -> pred_lcl_T:
        """Fit global step.

        Chooses a local probability peak prominence threshold for converting 
        local proabilities (a Series) to a collection of timestamps (event 
        predictions).

        Args:
            events_nested (Dict[str, pd.DataFrame]): A dictionary of sensor 
                data windows containing labeled events, keyed by deployment ID.
            nonevents_nested (Dict[str, pd.DataFrame]): As events_nested, but 
                windows contain non-events.
            sensors (Dict[str, pd.DataFrame]): Sensor data.
            events (Dict[str, pd.DatetimeIndex]): Labeled events.
            mask (Dict[str, np.ndarray]): Training mask.

        Returns:
            Dict[str, pd.Series]: The results of the cross-validation local 
                predictions. Used in the boosting step to find false positives.
        """
        # Global step (using internal cross validation)
        n_folds = min(self.n_folds, len(sensors.keys()))
        kf = selection.KFold(n_folds)
        deployids = np.array(list(sensors.keys()))
        local_proba = dict()
        for train_idx, test_idx in kf.split(deployids):
            event_train_X = pd.concat([v 
                                       for k, v in events_nested.items() 
                                       if k in deployids[train_idx]])
            nonevent_train_X = pd.concat([v 
                                          for k, v in nonevents_nested.items() 
                                          if k in deployids[train_idx]])
            train_X = event_train_X.append(nonevent_train_X)
            train_y = np.concatenate([np.full(len(event_train_X), 1.0), 
                                      np.full(len(nonevent_train_X), 0.0)])
            test_sensors = sb_util.filter_dict(sensors, deployids[test_idx])
            self._fit_local(train_X, train_y, clone=True)
            local_proba.update(
                self._predict_local(test_sensors, mask, clone=True)
            )
            
        prominence = np.linspace(0, 1, 25)
        f1 = np.zeros(prominence.shape)
        for i, p in enumerate(prominence):
            peaks = dict()
            for k, v in local_proba.items():
                idx, _ = signal.find_peaks(v.fillna(0), prominence=p)
                peaks[k] = (v, v.index[idx])
                
            outcomes = self.assess(peaks, events)
            tps = fps = fns = 0
            for o in outcomes.values():
                tps += (o == "TP").sum()
                fps += (o == "FP").sum()
                fns += (o == "FN").sum()
                
            f1[i] = tps / (tps + (fps + fns) / 2)
            
        self.prominence = prominence[np.argmax(f1)]
        return local_proba

    def _predict_global(self, local_proba: pred_lcl_T) -> pred_gbl_T:
        """Global step predictions.

        Args:
            local_proba (Dict[str, pd.Series]): A dict of local probability 
                predictions, keyed by deployment ID.

        Returns:
            Dict[str, pd.DatetimeIndex]: A dict of globabl predictions, i.e. 
                times of predicted events, keyed by deployment ID.
        """
        peaks = dict()
        for k, v in local_proba.items():
            idx, _ = signal.find_peaks(v.fillna(0), prominence=self.prominence)
            peaks[k] = v.index[idx]
            
        return peaks
        
    def _boost(self, 
               nonevents: nested_T, 
               sensors: sensors_T, 
               outcomes: outcomes_T) -> nested_T:
        fps = {d: o.index[o == "FP"] for d, o in outcomes.items()}
        nested_fps = sb_util.extract_nested(sensors, fps, self.win_size)
        boosted_nonevents = dict()
        for k in nonevents:
            if k in nested_fps: 
                boosted_nonevents[k] = pd.concat([nonevents[k], nested_fps[k]])
            else:
                boosted_nonevents[k] = nonevents[k]
                
        return boosted_nonevents
    


from matplotlib.figure import Figure as matplotlibFigure
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import plotly.graph_objects as go
from plotly.graph_objs._figure import Figure as plotlyFigure
from plotly.subplots import make_subplots
import stickleback.util as sb_util

from stickleback.types import *
from typing import Dict, Union

figure_T = Union[plotlyFigure, matplotlibFigure]

def plot_sensors_events(deployid: str, 
                        sensors: sensors_T, 
                        events: events_T, 
                        interactive=True) -> figure_T:
    """Plot sensor data and labeled events.

    Args:
        deployid (str): Deployment ID to plot.
        sensors (Dict[str, pd.DataFrame]): Sensor data.
        events (Dict[str, pd.DatetimeIndex]): Labeled events.
        interactive (bool, optional): Interaction flag. If True, produces an 
            interactive plotly figure. Else, a static matplotlib figure. 
            Defaults to True.

    Returns:
        Union[plotlyFigure, matplotlibFigure]: Figure with sensors and events.
    """
    _sensors = sensors[deployid]
    _events = _sensors.loc[events[deployid]]
    
    if interactive:
        return __plot_sensors_events_interactive(_sensors, _events)
    else:
        return __plot_sensors_events_static(_sensors, _events)

def __plot_sensors_events_interactive(sensors: pd.DataFrame, 
                                      event_sensors: pd.DataFrame) -> plotlyFigure:
    """Plot sensors and events interactively.

    Args:
        sensors (pd.DataFrame): Sensor data for one deployment.
        event_sensors (pd.DataFrame): Sensor data at event times for one 
            deployment.

    Returns:
        plotlyFigure: Interactive plotly figure.
    """
    fig = make_subplots(rows=len(sensors.columns), cols=1,
                        shared_xaxes=True,)
    for i, col in enumerate(sensors.columns):
        fig.append_trace(go.Scatter(
            x=sensors.index,
            y=sensors[col],
            mode="lines"
        ), row=i + 1, col=1)
        fig.append_trace(go.Scatter(
            x=event_sensors.index,
            y=event_sensors[col],
            mode="markers"
        ), row=i + 1, col=1)
        fig.update_yaxes(title_text=col, row=i + 1, col=1)
        
    fig.update_layout(showlegend=False)
    return fig

def __plot_sensors_events_static(sensors: pd.DataFrame, 
                                 event_sensors: pd.DataFrame) -> matplotlibFigure:
    """Plot sensors and events statically.

    Args:
        sensors (pd.DataFrame): Sensor data for one deployment.
        event_sensors (pd.DataFrame): Sensor data at event times for one 
            deployment.

    Returns:
        matplotlibFigure: Static matplotlib figure.
    """
    fig, axs = plt.subplots(len(sensors.columns), 1)
    for i, col in enumerate(sensors.columns):
        # sensor data
        axs[i].plot(sensors.index, sensors[col], "-", zorder=1)
        # events
        axs[i].scatter(event_sensors.index, 
                       event_sensors[col], 
                       facecolors="none", 
                       edgecolors="r", 
                       zorder=2)
        axs[i].set_ylabel(col)
        
    return fig

def plot_predictions(deployid: str, 
                     sensors: sensors_T, 
                     predictions: prediction_T,
                     outcomes: outcomes_T = None, 
                     interactive: bool = True) -> figure_T:
    """Plot model predictions.

    Args:
        deployid (str): Deployment ID to plot.
        sensors (Dict[str, pd.DataFrame]): Sensor data.
        predictions (Dict[str, Tuple[pd.Series, pd.DatetimeIndex]]): Model 
            predictions (as from Stickleback.predict()).
        outcomes (Dict[str, pd.Series], optional): Prediction outcomes (as from
            Stickleback.assess()). Defaults to None.
        interactive (bool, optional): Interaction flag. If True, produces an 
            interactive plotly figure. Else, a static matplotlib figure. 
            Defaults to True.

    Returns:
        Union[plotlyFigure, matplotlibFigure]: Figure with model predictions.
    """
    lcl, gbl = predictions[deployid]
    data = sensors[deployid].join(lcl)

    predicted_only = data.loc[gbl]
    actual_only = None
    if outcomes is not None:
        predicted_only = predicted_only.join(outcomes[deployid])
        is_actual = outcomes[deployid].isin(["TP", "FN"])
        actual_idx = outcomes[deployid].index[is_actual]
        actual_only = data.loc[actual_idx].join(outcomes[deployid])

    if interactive:
        return __plot_predictions_interactive(data, predicted_only, 
                                              actual_only)
    else:
        return __plot_predictions_static(data, predicted_only, actual_only)

def __plot_predictions_interactive(data: pd.DataFrame, 
                                   predicted: pd.DataFrame, 
                                   actual: pd.DataFrame = None) -> plotlyFigure:
    """Plot model predictions interactively.

    Args:
        data (pd.DataFrame): Sensor data for one deployment.
        predicted (pd.DataFrame): Sensor data at predicted event times.
        actual (pd.DataFrame, optional): Sensor data at actual event times.
            Defaults to None.

    Returns:
        plotlyFigure: Interactive plotly figure.
    """
    fig = make_subplots(rows=len(data.columns), cols=1,
                        shared_xaxes=True,)

    pred_color = "purple"
    if "outcome" in predicted.columns:
        pred_color = ["blue" if o == "TP" else "red" 
                      for o in predicted["outcome"]]
    if actual is not None:
        actual_color = ["blue" if o == "TP" else "red" 
                        for o in actual["outcome"]]

    for i, col in enumerate(data):
        # Line plot
        fig.append_trace(go.Scatter(
            x=data.index,
            y=data[col],
            mode="lines"
        ), row=i + 1, col=1)
        # Predicted events
        fig.append_trace(go.Scatter(
            x=predicted.index,
            y=predicted[col],
            marker_color=pred_color,
            mode="markers"
        ), row=i + 1, col=1)
        # Actual events
        if actual is not None:
            fig.append_trace(go.Scatter(
                x=actual.index,
                y=actual[col],
                mode="markers",
                marker_symbol="circle-open",
                marker_size=10,
                marker_color=actual_color,
            ), row=i + 1, col=1)
        fig.update_yaxes(title_text=col, row=i + 1, col=1)
        
    fig.update_layout(showlegend=False)
    return fig

def __plot_predictions_static(data: pd.DataFrame, 
                              predicted: pd.DataFrame, 
                              actual: pd.DataFrame = None) -> matplotlibFigure:
    """Plot model predictions statically.

    Args:
        data (pd.DataFrame): Sensor data for one deployment.
        predicted (pd.DataFrame): Sensor data at predicted event times.
        actual (pd.DataFrame, optional): Sensor data at actual event times. 
            Defaults to None.

    Returns:
        matplotlibFigure: Static matplotlib figure.
    """
    fig, axs = plt.subplots(len(data.columns), 1)
    for i, col in enumerate(data):
        # sensor data
        axs[i].plot(data.index, data[col], "-", zorder=1)
        axs[i].set_ylabel(col)
        # predicted events
        axs[i].scatter(predicted.index, 
                        predicted[col], 
                        c=["blue" if o == "TP" else "red" 
                           for o in predicted["outcome"]], zorder=2)
        # actual events
        axs[i].scatter(actual.index, 
                        actual[col], 
                        edgecolors=["blue" if o == "TP" else "red" 
                                    for o in actual["outcome"]],
                        facecolors="none",
                        zorder=3)
        
    return fig

def outcome_table(outcomes: outcomes_T,
                  sensors: sensors_T) -> pd.DataFrame:
    """Create a table of model prediction outcomes.

    Args:
        outcomes (Dict[str, pd.Series]): Model prediction outcomes (as from 
            Stickleback.assess()).
        sensors (Dict[str, pd.DataFrame]): Sensor data.

    Returns:
        pd.DataFrame: A table of prediction outcomes. Each row is a deployment.
            Columns are F1 score (F1); number of true positives (TP), false 
            positives (FP), and false negatives (FN); and deployment duration 
            in hours (Duration (hours)).
    """
    def add_deployid(d, o):
        result = pd.DataFrame(o)
        result.insert(0, "deployid", np.full(len(o), [d]))
        return result
    
    counts = pd.concat([add_deployid(d, o) for d, o in outcomes.items()])
    result = (counts
              .groupby(["deployid", "outcome"])
              .size()
              .unstack(fill_value=0))
    
    for o in ["TP", "FP", "FN"]:
        if o not in result:
            result[o] = 0
    
    result["F1"] = sb_util.f1(result["TP"], result["FP"], result["FN"])
    
    dur_dict = {k: (v.index[-1] - v.index[0]) / pd.Timedelta(hours=1) 
                for k, v in sensors.items()}
    durations = pd.DataFrame.from_dict(dur_dict, "index")
    durations.columns = ["Duration (hours)"]
    result = result.join(durations)
    
    result = result[["F1", 
                     "TP", 
                     "FP", 
                     "FN", 
                     "Duration (hours)"]]
    return result



import pandas as pd
import pickle
import numpy as np
import sktime.datatypes._panel._convert as convert 
from stickleback.types import *
from typing import Collection, Dict, Tuple, Union

def extract_nested(sensors: sensors_T, 
                   idx: Dict[str, pd.DatetimeIndex], 
                   win_size: int) -> nested_T:
    """Extract windows from sensor data in sktime nested DataFrame format.

    Args:
        sensors (Dict[str, pd.DataFrame]): Sensor data.
        idx (Dict[str, pd.DatetimeIndex]): Dict of indicies, keyed by 
            deployment ID. Each output window will be centered on one of these
            indices.
        win_size (int): Window size.

    Returns:
        Dict[str, pd.DataFrame]: Dict of data windows in sktime nested 
            DataFrame format.
    """
    win_size_2 = int(win_size / 2)

    def _extract(_deployid: str, _idx: pd.DatetimeIndex):
        _sensors = sensors[_deployid]
        idx = _sensors.index.get_indexer(_idx)
        data_3d = np.empty([len(idx), len(_sensors.columns), win_size], float)
        data_arr = _sensors.to_numpy().transpose()
        start_idx = idx - win_size_2
        for i, start in enumerate(start_idx):
            data_3d[i] = data_arr[:, start:(start + win_size)]
        nested = convert.from_3d_numpy_to_nested(data_3d)
        nested.columns = _sensors.columns
        nested.index = _sensors.index[idx]
        return nested

    return {d: _extract(d, i) for d, i in idx.items()}

def extract_all(sensors: sensors_T, 
                nth: int, 
                win_size: int, 
                mask: mask_T = None) -> nested_T:
    """Extract all windows from sensor data in sktime nested DataFrame format.

    Args:
        sensors (Dict[str, pd.DataFrame]): Sensor data.
        nth (int): Sliding window step size.
        win_size (int): Window size.
        mask (mask_T, optional): Window mask. Only windows in sensors[
            deployid] where mask[deployid] is True will be extracted. If None, 
            no mask applied. Defaults to None.

    Returns:
        Dict[str, pd.DataFrame]: Dict of data windows in sktime nested 
            DataFrame format.
    """
    if mask is None:
        mask = {d: np.full(len(sensors[d]), True) for d in sensors}
        
    win_size_2 = int(win_size / 2)
    idx = dict()
    for d in sensors:
        _idx = np.arange(win_size_2, len(sensors[d]) - win_size_2, nth)
        # Next line (admittedly) confusing. Look up _idx in mask[d] and keep
        # only those where mask is True
        _idx = _idx[mask[d][_idx]]
        idx[d] = sensors[d].index[_idx]
        
    return extract_nested(sensors, idx, win_size)
    
def sample_nonevents(sensors: sensors_T, 
                     events: events_T, 
                     win_size: int, 
                     mask: mask_T = None, 
                     seed: int = None) -> nested_T:
    """Randomly sample non-event windows from sensor data.

    Args:
        sensors (Dict[str, pd.DataFrame]): Sensor data.
        events (Dict[str, pd.DatetimeIndex]): Labeled events.
        win_size (int): Window size.
        mask (Dict[str, np.ndarray], optional): Window mask. Only windows in 
            sensors[deployid] where mask[deployid] is True will be extracted. 
            If None, no mask applied. Defaults to None.
        seed (int, optional): Random number generator seed. Defaults to None.

    Returns:
        Dict[str, pd.DataFrame]: Dict of data windows in sktime nested 
            DataFrame format.
    """
    win_size_2 = int(win_size / 2)
    rg = np.random.Generator(np.random.PCG64(seed))
    if mask is None:
        mask = {d: np.full(len(sensors[d]), True) for d in sensors}

    def _diff_from(xs: np.ndarray, ys: np.ndarray) -> np.ndarray:
        return np.array([np.min(np.abs(x - ys)) for x in xs])

    def _sample(_sensors: pd.DataFrame, 
                _events: pd.DatetimeIndex, 
                _mask: np.ndarray):
        nonevent_choices = np.arange(win_size_2, 
                                     len(_sensors) - win_size_2, 
                                     win_size)
        nonevent_choices = nonevent_choices[_mask[nonevent_choices]]
        diff_from_event = _diff_from(nonevent_choices, 
                                     _sensors.index.searchsorted(_events))
        nonevent_choices = nonevent_choices[diff_from_event > win_size]
        return _sensors.index[rg.choice(nonevent_choices, 
                                        size=len(_events), 
                                        replace=True)]

    idx = {d: _sample(sensors[d], events[d], mask[d]) for d in sensors}
    return extract_nested(sensors, idx, win_size)

def align_events(events: events_T, sensors: sensors_T) -> events_T:
    """Align labeled events with sensor data indices.

    Args:
        events (Dict[str, pd.DatetimeIndex]): Labeled events
        sensors (Dict[str, pd.DataFrame]): Sensor data

    Returns:
        Dict[str, pd.DatetimeIndex]: As events, but times shifted to closest 
            index in sensors.
    """
    return {d: sensors[d].index[sensors[d].index.searchsorted(e)] 
            for d, e in events.items()}

def filter_dict(d: Dict, ks: Collection) -> Dict:
    """Filter a dict by keys.

    Args:
        d (Dict): A dictionary.
        ks (Collection): Keys to keep.

    Returns:
        Dict: d, but with only the keys in ks.
    """
    return {k: v for k, v in d.items() if k in ks}

def save_fitted(sb: "Stickleback", 
                fp: str,
                sensors: sensors_T = None, 
                events: events_T = None, 
                mask: mask_T = None) -> None:
    """Save a fitted Stickleback model.

    Args:
        sb (Stickleback): Stickleback model.
        fp (str): File path.
        sensors (Dict[str, pd.DataFrame], optional): Sensor data used to fit 
            model. Defaults to None.
        events (Dict[str, pd.DatetimeIndex], optional): Labeled events used to 
            fit model. Defaults to None.
        mask (Dict[str, np.ndarray], optional): Mask used to fit model. 
            Defaults to None.
    """
    objects = (sb, sensors, events, mask,)
    with open(fp, 'wb') as f:
        pickle.dump(objects, f)
        
def load_fitted(fp: str) -> Tuple["Stickleback", 
                                  sensors_T, 
                                  events_T, 
                                  mask_T]:
    """Load a fitted Stickleback model.

    Returns:
        A tuple containing:
            The fitted Stickleback model
            The sensor data used to fit the model (possibly None).
            The labeled events used to fit the model (possibly None).
            The mask used to fit the model (possibly None).
    """
    with open(fp, 'rb') as f:
        result = pickle.load(f)
    
    return result

def f1(tps: Union[float, pd.Series],
       fps: Union[float, pd.Series],
       fns: Union[float, pd.Series]) -> Union[float, pd.Series]:
    """Calculate F1 score

    Args:
        tps (Union[float, pd.Series]): Count(s) of true positives.
        fps (Union[float, pd.Series]): Count(s) of false positives.
        fns (Union[float, pd.Series]): Count(s) of false negatives.

    Returns:
        Union[float, pd.Series]: F1 score(s).
    """
    return tps / (tps + (fps + fns) / 2)



import pathlib
from setuptools import setup

README = (pathlib.Path(__file__).parent / "README.md").read_text()

setup(
    name="stickleback",
    version="0.1.2",
    description="Automated behavioral event detection in bio-logging data.",
    long_description=README,
    long_description_content_type="text/markdown",
    url="http://github.com/FlukeAndFeather/stickleback",
    author="Max Czapanskiy",
    author_email="maxczapanskiy@gmail.com",
    license="MIT",
    packages=['stickleback'],
    include_package_data=True,
    package_data={'stickleback': ['data/*']},
    install_requires=[
            "jupyter~=1.0",
            "matplotlib~=3.4",
            "netcdf4~=1.5",
            "numpy~=1.20",
            "pandas~=1.2",
            "plotly~=4.12",
            "scikit-learn~=0.24",
            "scipy~=1.6",
            "sktime~=0.8"
        ]
)
